//Developer:	Qishi Wu
//Date:			June 2004
//Place:		CSMD @ ORNL

#ifndef _UDPCLI_H_
#define _UDPCLI_H_

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include "ioutil.h"
#include <pthread.h>

/* IMPORTANT: needs to calibrated per system */
#define BETA 2.00

//#define SERVER_IP	"160.91.77.77"
//#define SERVER_IP	"160.91.77.111"  //ozy2.csm.ornl.gov
//#define SERVER_IP	"130.39.224.177"  //resource.rrl.lsu.edu
//#define SERVER_IP	"192.31.96.56"	//unet2-mgmt.csm.ornl.gov
#define	SERVER_IP	"192.31.96.202"	//unet2.csm.ornl.gov -- Gige NIC

#define SERVER_PORT	7058



#define BANDWIDTH	10000.0	//Mbps

#define MTU			1500	//bytes
#define HEADER		4		//sizeof(int): dgSeqNo
#define DGSIZE		(MTU-28) //1500- 20 (IP header) - 8 (UDP header) = 1472 bytes
#define DATASIZE	(DGSIZE-HEADER)


//sending rate control parameters
#define CWIN		1		// datagrams
#define SLEEP		10		// microseconds
#define SENDRATE	100.0	// sending rate in Mbps

typedef struct FileInfoNode
{
	char pFileName[100];
	bool bIsADir;
	struct FileInfoNode *pNext;

} FileInfoNode;

typedef struct UDPParams
{
     int sockfd_tcp;
     int sockfd_udp;
     struct sockaddr_in servaddr;
     socklen_t servlen;
     int cwin;
     int *sleep_us;
     char *pFileName;
     INT64 fileSize_bytes;
     int dgSize;
     int totalDgsInFile;
     bool *bComplete;
     int  *rexmtCount;
     bool *bAllSent;
} UDPParams;

typedef struct RexmtParams
{
     int sockfd_udp;
     struct sockaddr_in servaddr;
     socklen_t servlen;
     int sleep_us;
     int numRexmt;
     char *pRecvLine;
     int dgSize;

} RexmtParams;

void PrintCommand();
void ReadFileSizeInBytes(char *pFileName, INT64 &fileSize_bytes);

void ParameterParser(int argc, char **argv, char *pSvrAddr, int *pSvrPort,
					 int *pMTU, int *pcwin, int *psleep_us, double *pReqBW, 
					 char *pFileName);

/*void ParameterParser(int argc, char **argv, char *pSvrAddr, int *pSvrPort,
					 int *pMTU, int *pcwin, int *psleep_us, double *pReqBW,
					 INT64 *pfileSize, char *pRequestName);*/

//void SendUDPDatagrams(int sockfd_tcp, int sockfd_udp, struct sockaddr_in servaddr, socklen_t servlen, int cwin, int sleep_us, char *pFileName, INT64 fileSize_bytes, int dgSize, int totalDgsInFile);

void *Rexmt(void *arg);

void *SendUDPDatagrams(void *arg);
	
int ThreeWayHandshakeAtSender(int sockfd_udp, struct sockaddr_in servaddr, socklen_t servlen,
							   int cwin, int *sleep_us, char *pFileName, INT64 fileSize, int dgSize, int totalNumOfDgs, double &rtt);

void ReceiveAcknowledge(char *pLine, int sockfd_tcp);

bool IsFileListReady(char *pFileListName);

void CreateFileList(FileInfoNode **ppHead, char *pRequestName);
void DeleteFileList(FileInfoNode *pHead);


int getSleepFromRate(double dgSize, double reqBW);


int getRateFromSleep(double dgSize, double sleep_us);


#endif

