//Developer:	Qishi Wu + Ben Eckart
//Date:			June 2004 + June 2007
//Place:		CSMD @ ORNL + TTU

#include "ioutil.h"
#include "udpserver.h"

//#define LATENCY_SIM

/*******************************************
 * 	Global variables and mutexes
 ******************************************/

/* Array to hold dropped datagrams.  */
bool *pDroppedDgList;

#ifdef LATENCY_SIM
int latency = -1;
int MAX_LATENCY_SIZE = 500000;
char **datagramLatency;
int *dgLengthLatency;
struct timeval *timestampLatency;
#endif

/* Index used for datagram memory management */
int glb_index = 0;


/* Counter for number of times recv call has been made.
 * Used in profiling recv thoughput for rate control. */
int glb_recvCount = 0; 

/* Counter used in calculating disk throughput. */
int glb_diskCount = 0; 

/* Mutex for set pDroppedDgList. */
pthread_mutex_t mutexRexmt = PTHREAD_MUTEX_INITIALIZER;

/* Mutex for synchronizing multiple disk threads. */
pthread_mutex_t mutexDisk = PTHREAD_MUTEX_INITIALIZER;

//pthread_mutex_t mutexDiskDgCount = PTHREAD_MUTEX_INITIALIZER;
//pthread_mutex_t mutexDgCount = PTHREAD_MUTEX_INITIALIZER;

int main(int argc, char **argv)
{
	/* Socket connections */
	int		svrPort = SERVER_PORT;
	int		sockfd_tcp_listen;
	int		sockfd_udp, connfd_tcp;
	struct 	sockaddr_in servaddr, cliaddr;
	socklen_t clilen = sizeof(cliaddr);
	if(argc == 3 && strncmp(argv[1], "-p", 2) == 0)
		svrPort = atoi(argv[2]);

#ifdef LATENCY_SIM
	else if(argc == 5 && strncmp(argv[3], "-l", 2) == 0)
		latency = atoi(argv[4]);
#endif

	else if(argc != 1)
	{
		printf("\nCommand format:\nudpserver [-p portNo] [-l latency(us)]\n");
		exit(0);
	}

	sockfd_udp = socket(AF_INET, SOCK_DGRAM, 0);
	sockfd_tcp_listen = socket(AF_INET, SOCK_STREAM, 0);
		
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons((unsigned short)svrPort);

	if(bind(sockfd_udp, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
	{
		printf("\nbind sockfd_udp error!\n");
		exit(0);
	}

	if(bind(sockfd_tcp_listen, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
	{
		printf("\nbind sockfd_tcp_listen error!\n");
		exit(0);
	}

	listen(sockfd_tcp_listen, 5);
	while(true)
	{
		
		printf("\nTCP server is listening at port %d ...\n", svrPort);

		/* accept is blocking */
		connfd_tcp = accept(sockfd_tcp_listen, (struct sockaddr *)&cliaddr, &clilen);

		startTransfer(sockfd_udp, connfd_tcp);
	}
}

int startTransfer(int sockfd_udp, int connfd_tcp)
{	
	/* Datagram processing & sending */
	bool	bAllSent;
	bool	b1stDatagram;
	bool 	*pDgChkList; /* will be totalDgsInFile large */
	int		dgID;
	int		mtu, dgSize, rate, totalDgsInFile;
	int		cwin, cwinInOneRTT;
	int		dgNoIndex, dgCount;
  	int		numRecv, dataSize, numRecvGoodDgs;
	int		numDupDgs;
	int 	totalDgsRxmt;
	int		*pDgSeqNoList; /* will be totalDgsInFile large */
	INT64	totalRecvSoFar, recvGoodBytes;
	double	RTTms;
	char 	*pDatagram; /* will be dgSize large */
	char	pAckLine[MAX_LINE_LEN];
	char	pRecvLine[MAX_LINE_LEN];
	
 	/* Counters */
	int		i, j;
 	
	/* File processing */
	char	pFileName[MAX_STRING_LEN], pTempFileName[MAX_STRING_LEN]; 
	INT64	fileSize;
	FILE	*fpData;
	int		randomNo;
	char	pDir[MAX_STRING_LEN], pTempLine[MAX_STRING_LEN];

	/* Timing */
	struct 	timeval	timeCur;
	long int startTime, endTime, timeInterval;
	double	duration; //in seconds
	
	/* Statistics */
	double	receivingRate;
	double 	rxmtRate;
	double	targetRate;

	/* Memory Management */
	int 	reallocCount;
	bool 	isNew;
	int		*pSeqNewDgs;
	int		seqNewDgsIndex;
	
	/* File writing */
	bool	bDiskSleep = true;

	/* Primary data structure which holds received datagrams. */
	char	**ppDatagrams;

	/* Threads */
	pthread_t postProcessThreadID;
	pthread_t rexmtThread;
	pthread_t rateControlThread;
	pthread_t diskThread2;
	pthread_t diskThread;
	pthread_attr_t *tPriority;

	numRecv = ReadLine(connfd_tcp, pRecvLine, MAX_LINE_LEN);

	if(numRecv > 0 && strncmp(pRecvLine, "FILE MTU", 8) == 0)
	{
		sscanf(pRecvLine, "FILE MTU %d", &mtu);
		dgSize = mtu - 28;

		ThreeWayHandshakeAtReceiver(sockfd_udp, dgSize, &cwin, &rate, pFileName, &fileSize, &totalDgsInFile, &RTTms);

		printf("\nPrepare to receive file %s (%lld bytes)", pFileName, fileSize);
		fflush(stdout);

		/* initialization */
		b1stDatagram = true;	
		bAllSent = false;
		bDiskSleep = true;	
		numRecvGoodDgs = 0;
		recvGoodBytes = 0;
		dgCount = 0;
		totalRecvSoFar = 0;
		totalDgsRxmt = 0;
		numDupDgs = 0;
		dgNoIndex = 0;
		glb_diskCount = 0;
		glb_recvCount = 0;
		glb_index = 0;

		pDroppedDgList = (bool *)malloc(sizeof(bool)*totalDgsInFile);
		pSeqNewDgs = (int *)malloc(totalDgsInFile*sizeof(int));
		seqNewDgsIndex = 0;
		reallocCount = 0;
		isNew = false;
		pDgChkList = (bool *)malloc(sizeof(bool)*totalDgsInFile);
		pDgSeqNoList = (int *)malloc(sizeof(int)*totalDgsInFile);
		ppDatagrams = (char **)malloc(sizeof(char*)*totalDgsInFile);

#ifdef LETENCY_SIM

		if(latency != -1)
		{
			datagramLatency = (char **)malloc(sizeof(char*)*MAX_LATENCY_SIZE);
			dgLengthLatency = (int *)malloc(sizeof(int)*MAX_LATENCY_SIZE);
		timestampLatency = (struct timeval*)malloc(sizeof(struct timeval)*MAX_LATENCY_SIZE);
			for(i = 0; i < MAX_LATENCY_SIZE; i++)
	          dgLengthLatency[i] = -1;
        }
#endif

		/* initialize check list */
		for(j = 0; j < totalDgsInFile; j ++)
		{
			pDgChkList[j] = false;
			ppDatagrams[j] = NULL;
			pDroppedDgList[j] = false;
		}		

		/* calculate target rate if sleep explicitly set */
		//if(sleep != 0)
			//targetRate = ( (dgSize*8) /(double)sleep ) - ( (2*dgSize*8) /( (double)sleep*sleep ) );
		//else
		
		targetRate = rate;
			
		printf("Target Rate: %f\n", targetRate);

		strncpy(pTempFileName, pFileName, strlen(pFileName));
		pTempFileName[strlen(pFileName)] = '\0';
		for(i = strlen(pFileName)-1; i >= 0; i --)
		{
			if(pFileName[i] == '/')
			{
				strncpy(pTempFileName+i+1, ".tmp", 5);
				break;
			}
			if(i == 0)
				strncpy(pTempFileName, ".tmp", 5);
		}
		srand((unsigned)time(NULL));
		while(1)
		{
			randomNo = rand();
			sprintf(pTempFileName + strlen(pTempFileName), "%d", randomNo);
			if((fpData = (FILE *)fopen64(pTempFileName, "w")) != NULL)
				break;
		}

		/* Attempt to set unbuffered Disk I/O.
		 * In reality, this has little impact on latencies
		 * when synchronous disk activity is enforced through
		 * fsync calls. */
		if(setvbuf(fpData, NULL, _IONBF, 0) != 0)
			printf("\nERROR: Setting unbuffered I/O\n");
		else
			printf("\nSUCCESS: Setting unbuffered I/O\n");

		/* Attempt to modify system buffer */
		if(ModifyUDPSystemBuffer(RECV_BUF_SIZE, sockfd_udp) < RECV_BUF_SIZE)
			printf("\nWARNING: UDP buffer not increased, performance will suffer!");
		
		tPriority = (pthread_attr_t *)malloc(sizeof(pthread_attr_t));
		SetThreadPriority(99, tPriority);

		//set up disk thread
		DiskThreadInfo *pDiskThreadInfo = (DiskThreadInfo *)malloc(sizeof(DiskThreadInfo));
		pDiskThreadInfo->dgSize = dgSize;
		pDiskThreadInfo->totalDgsInFile = totalDgsInFile;
		pDiskThreadInfo->fpData = fpData;
		pDiskThreadInfo->bDiskSleep = &bDiskSleep;
		pDiskThreadInfo->bAllSent = &bAllSent;
		pDiskThreadInfo->ppDatagrams = ppDatagrams;
		if(pthread_create(&diskThread, tPriority, DiskThread, pDiskThreadInfo) != 0)
		{
			printf("\nError: cannot create file-processing thread!\n\n");
			fflush(stdout);
			exit(0);
		}

		//set up disk thread
		DiskThreadInfo *pDiskThreadInfo2 = (DiskThreadInfo *)malloc(sizeof(DiskThreadInfo));
		pDiskThreadInfo2->dgSize = dgSize;
		pDiskThreadInfo2->totalDgsInFile = totalDgsInFile;
		pDiskThreadInfo2->fpData = fpData;
		pDiskThreadInfo2->bDiskSleep = &bDiskSleep;
		pDiskThreadInfo2->bAllSent = &bAllSent;
		pDiskThreadInfo2->ppDatagrams = ppDatagrams;
		if(pthread_create(&diskThread2, tPriority, DiskThread, pDiskThreadInfo2) != 0)
		{
			printf("\nError: cannot create file-processing thread!\n\n");
			fflush(stdout);
			exit(0);
		}

		//set up retransmit thread
		RexmtThreadInfo *pRexmtThreadInfo = (RexmtThreadInfo *)malloc(sizeof(RexmtThreadInfo));
		pRexmtThreadInfo->connfd_tcp = connfd_tcp;
		pRexmtThreadInfo->pTotalDgsRxmt = &totalDgsRxmt;
		pRexmtThreadInfo->pbAllSent = &bAllSent;
		pRexmtThreadInfo->pbNotDone = &bDiskSleep;
		pRexmtThreadInfo->totalDgsInFile = totalDgsInFile;
				
		if(pthread_create(&rexmtThread, NULL, RexmtThread, pRexmtThreadInfo) != 0)
		{
			printf("\nError: cannot tcp retransmitting thread!\n\n");
			fflush(stdout);
			exit(0);
		}
        
#ifdef LATENCY_SIM
		if(latency != -1)
		{
			pthread_t writeLatThread;
			if(pthread_create(&writeLatThread, NULL, writeLatencyThread, &connfd_tcp) != 0)
			{
				printf("\nError: cannot tcp retransmitting thread!\n\n");
				fflush(stdout);
				exit(0);
			}
		}
#endif
        
		if(fileSize >= MAX_MEMORY*ONE_MILLION)
		{
RateControlThreadInfo *pTRTI = (RateControlThreadInfo *)malloc(sizeof(RateControlThreadInfo));
			pTRTI->sleep = BW_CONTROL_USLEEP;
			pTRTI->dgSize = dgSize;
			pTRTI->targetRate = (int)targetRate;
			pTRTI->connfd_tcp = connfd_tcp;
			pTRTI->fileSize = fileSize;
			pTRTI->bAllSent = &bAllSent;
			pTRTI->bFirstDg = &b1stDatagram;

			if(pthread_create(&rateControlThread, NULL, RateControlThread, pTRTI) != 0)
			{
				printf("\nError: cannot start control thread!\n\n");
				fflush(stdout);
				exit(0);
			}
		}
		
		//notify client to start transfer
		strncpy(pTempLine, "SVRREADY\n", 9);
		writen(connfd_tcp, pTempLine, 9);

		printf("\nStart receiving ... \n", pFileName);
	}
	else if(numRecv > 0 && strncmp(pRecvLine, "DIR ", 4) == 0)
	{
		/*printf("DIRECTORY\n");
		char *dirString = new char[1000];
		char *sysString = new char[1000];
		strcpy(sysString, "mkdir ");
		sscanf(pRecvLine, "DIR %s", dirString);
		system(strcat(sysString,dirString));*/
		//not implemented
	}
	else
	{
		printf("\nIncorrect client request! ", pFileName);
		exit(0);
	}
	
	/* Sending loop */
	for(; ;)
	{
		
		/* allocate memory for new datagram */
		if(reallocCount >= glb_diskCount)
		{
			pDatagram = (char *)malloc(sizeof(char)*dgSize);
			isNew = true;
		}
		else
		{
			pDatagram = (char *)realloc(ppDatagrams[reallocCount++],dgSize);
			glb_index++;
		}
		
		/* recv blocks execution when udp buffer is empty */
		numRecv = recv(sockfd_udp, pDatagram, dgSize, 0);

		/* update global count for rate control thread */
		//pthread_mutex_lock(&mutexDgCount);
		glb_recvCount++;
		//pthread_mutex_unlock(&mutexDgCount);

		/* check for possible errors */
		if(numRecv == -1 || numRecv == 0)
		{
			printf("\nFile receiving: recv error.\n");

			exit(0);
		}
		
		/* if first datagram received, start recording time
		 * for throughput and rexmt rate calculations. */
		if(b1stDatagram)
		{
			b1stDatagram = false;
			gettimeofday(&timeCur, NULL);
			startTime = timeCur.tv_usec + timeCur.tv_sec * 1000000;	
		}

		/* extract datagram ID from received datagram */
        memcpy(&dgID, pDatagram, 4);
		dgID = ntohl(dgID);

		/* if in error, assume corrupted, try to ignore */
        if(dgID < 0 || dgID > totalDgsInFile - 1)
        {	
        	printf("\nread a bad dgID: %d\n", dgID);
			fflush(stdout);
			continue;
        }

		/* received the last datagram, but this doesn't 
		 * always mean we've received everything yet */
		if(dgID == totalDgsInFile - 1)
		{
			bAllSent = true;
		}

		dataSize = numRecv - 4;
		totalRecvSoFar += (INT64)(numRecv + 28);
		dgCount ++;

		// it is a new dg
        if(!pDgChkList[dgID])	
        {
			pthread_mutex_lock(&mutexRexmt);
			pDroppedDgList[dgID] = false;
			pthread_mutex_unlock(&mutexRexmt);
        	pDgChkList[dgID] = true;
            numRecvGoodDgs ++;
			recvGoodBytes += (INT64)dataSize;

			ppDatagrams[dgNoIndex] = pDatagram;
			if(isNew)
			{
				pSeqNewDgs[seqNewDgsIndex++] = dgNoIndex;	
				isNew = false;
			}

			pDgSeqNoList[dgNoIndex] = dgID;
			dgNoIndex ++;

			// check for missed datagrams, 
			// update dropped list appropriately
			if(dgID - pDgSeqNoList[dgNoIndex-2] > 1)
			{
				pthread_mutex_lock(&mutexRexmt);

				for(i = pDgSeqNoList[dgNoIndex-2] + 1; i < dgID; i++)
				{
					if(!pDgChkList[i])
						pDroppedDgList[i] = true;
				}	
				pthread_mutex_unlock(&mutexRexmt);
			}
		}
	    else
	    {
			// we have a dupe
	       	numDupDgs ++;
			--reallocCount; 
	    }
		
		//we are finished!			
		if(dgNoIndex == totalDgsInFile)
		{
			bDiskSleep = false;
			
			gettimeofday(&timeCur, NULL);
			endTime = timeCur.tv_usec + timeCur.tv_sec * 1000000;
			duration = (double) (endTime - startTime) / 1000000.0;
			receivingRate = (double)totalRecvSoFar * 8.0 / duration / 1000000.0;  // rate: Mbps
			rxmtRate =  (double)(totalDgsRxmt) / (double)dgCount * 100.0;

			sprintf(pAckLine, "ACK %d %d %lld %d %lf %lf\n", sleep, cwin, totalRecvSoFar, dgCount, receivingRate, rxmtRate);
			writen(connfd_tcp, pAckLine, strlen(pAckLine));
			usleep(100000);
			sprintf(pAckLine, "ACK %d %d %lld %d %lf %lf\n", sleep, cwin, totalRecvSoFar, dgCount, receivingRate, rxmtRate);
			writen(connfd_tcp, pAckLine, strlen(pAckLine));
			usleep(100000);
			close(connfd_tcp);

			printf("-->transfer completed!");
			printf("\nReceiver statistics:");
			printf("\nDuration: %lf",duration);
			printf("\nTotal recv dgs: %d, Total recv bytes: %lld", dgCount, totalRecvSoFar);
			printf("\nRecv good dgs: %d, Recv good bytes: %lld, Recv duplicate dgs: %d", numRecvGoodDgs, recvGoodBytes, numDupDgs);
			printf("\nrecv rate (throughput): %8.2lf Mbps, rxmt rate: %lf%%", receivingRate, rxmtRate);
			printf("\nrecv rate (goodput): %8.2lf Mbps",(double)fileSize * 8.0 / duration / 1000000.0);
			printf("\nDuration: %lf",duration);

			printf("\nI'm ready to receive the next file ...");
			printf("\n----------------------------------------\n");
			fflush(stdout);

			//wait for disk to finish writing contents of RAM
			pthread_join(diskThread, NULL);
			pthread_join(diskThread2,NULL);

			//flushes, then closes all open files
			fcloseall();

			printf("Starting file post-processing...");
			fflush(stdout);
			
			//////////////////Post File Process/////////////////////
			FileInfo *pFileInfo = (FileInfo *)malloc(sizeof(FileInfo));
			pFileInfo->dgSize = dgSize;
			pFileInfo->fileSize = fileSize;
			pFileInfo->totalDgsInFile = totalDgsInFile;
			pFileInfo->myHeaderSize = sizeof(int);
			strncpy(pFileInfo->pFileName, pFileName, strlen(pFileName));
			pFileInfo->pFileName[strlen(pFileName)] = '\0';
			strncpy(pFileInfo->pTempFileName, pTempFileName, strlen(pTempFileName));
			pFileInfo->pTempFileName[strlen(pTempFileName)] = '\0';
			pFileInfo->pDgSeqNoList = (int *)malloc(sizeof(int)*totalDgsInFile);
			memcpy(pFileInfo->pDgSeqNoList, pDgSeqNoList, totalDgsInFile*(pFileInfo->myHeaderSize));
			
			if(pthread_create(&postProcessThreadID, NULL, FileProcessingThread, pFileInfo) != 0)
			{
				printf("\nError: cannot create post file process thread!\n\n");
				fflush(stdout);
				exit(0);
			}
			/////////////////////////////////////////////////////////

			//wait for it to end
			pthread_join(postProcessThreadID, NULL);
			free(pDgSeqNoList);
			free(pDgChkList);
			

			//free up memory-- be careful, some elements in ppDatagrams
			//point to the same memory address due to realloc usage
			for(j = 0; j < seqNewDgsIndex; j++)
				free(ppDatagrams[pSeqNewDgs[j]]);

			//printf("\nTOTAL FREE CALLS: %d\n", j);
			//printf("\nREALLOC'ed DGS: %d\n", glb_index);
			
			free(ppDatagrams);
			free(pSeqNewDgs);
#ifdef LATENCY_SIM
			free(datagramLatency);
			free(dgLengthLatency);
			free(timestampLatency);
#endif
			break;		
		}
	} // end of for: infinite loop
	
	return 0;
}

void *RateControlThread(void *arg)
{
	RateControlThreadInfo *pArgs = (RateControlThreadInfo *)arg;
	int sleep = (int)pArgs->sleep;
	int dgSize = (int)pArgs->dgSize; //bytes
	int targetRate = (int)pArgs->targetRate;
	int connfd_tcp = (int)pArgs->connfd_tcp;
	INT64 fileSize = (INT64)pArgs->fileSize; //in bytes
	bool *bAllSent = (bool *)pArgs->bAllSent;
	bool *bFirstDg = (bool *)pArgs->bFirstDg;
	int maxMemory = MAX_MEMORY*1000000 - (1.0*sizeof(char *)/dgSize)*fileSize; //bytes
	
	//for slow recv throughput
	int prevDgCount = 0, dgCount = 0;
	struct timeval begin, end;
	double rate = 0.0, prevRate = targetRate;
	int goodCount = 0;
	char *packet = (char *)malloc(sizeof(char)*dgSize);
	double duration;
	

	//for slow disk write throughput
	double diskRate = 0.0, diskReqBW = 0.0;
	int prevDiskDgCount = 0, diskDgCount = 0;
	
	INT64 memLeft;
	INT64 bytesLeft; 
	
	double newRate;

	int dgBits = 8*dgSize;
	int dataBits = 8*(dgSize-4);

	double dgBitsPerTime = dgBits / 1000000.0;
	
	//sleep a little until the first dg is recv'ed
	while(*bFirstDg)
		usleep(5000);
	
	while(!*bAllSent)
	{

		//capture the global values
		prevDgCount = glb_recvCount;
		prevDiskDgCount = glb_diskCount;

		gettimeofday(&begin, NULL);
		usleep(sleep);
		gettimeofday(&end, NULL);
		
		//capture the global values
		dgCount = glb_recvCount;
		diskDgCount = glb_diskCount;
		
		//calculate recv rates based on recv'ed dgs
		duration = (end.tv_usec-begin.tv_usec) + (end.tv_sec-begin.tv_sec)*1000000; //seconds
		rate = (dgCount-prevDgCount)*dgBits/duration;
		
		//check to see if rate matches what was requested
		//if not, this could be a sign of slow packet processing = lost packets
		if(/*rate < prevRate - 100*/false)		
		{
			newRate = rate;
			printf("Throttling to %f for CPU\n", rate);
			//this causes oscillation when the performance is very low
		}
		else
		{
			//we need to capture our disk rate and set appropriately

			//how could this ever go up?
			memLeft = maxMemory - (dgCount-diskDgCount)*dgSize;

			bytesLeft = (fileSize/(dgSize-4) - (INT64)dgCount)*(dgSize-4);
			diskRate = (diskDgCount-prevDiskDgCount)*dataBits/duration;
			//diskRate = (diskDgCount-prevDiskDgCount)*dgBitsPerTime;
			
			newRate = (diskRate * bytesLeft) / (bytesLeft - memLeft);

			if(newRate > targetRate || memLeft > bytesLeft)
				newRate = targetRate;
			
			printf("diskR: %f newR: %f recvR: %f memLeft: %d bytLeft: %d\n",diskRate, newRate, rate, memLeft/1000000, bytesLeft/1000000);
			printf("%d %d\n", prevDgCount, dgCount);	
			//printf("diskR: %f newR: %f memLeft: %d12 bytLeft: %lld\n",diskRate, newRate, memLeft, bytesLeft);
		}
		
			
		prevRate = newRate;

		sprintf(packet, "RATE %d", (int)newRate);
		writen(connfd_tcp, packet, 5 + sizeof(int));
	}

}


void *RexmtThread(void *arg)
{
	RexmtThreadInfo *pRexmtThreadInfo = (RexmtThreadInfo *)arg;
	int connfd_tcp = (int)pRexmtThreadInfo->connfd_tcp;
	int *pTotalDgsRxmt = (int *)pRexmtThreadInfo->pTotalDgsRxmt;
	bool *pbAllSent = (bool *)pRexmtThreadInfo->pbAllSent;
	bool *pbNotDone = (bool *)pRexmtThreadInfo->pbNotDone;
	int	totalDgsInFile = (int)pRexmtThreadInfo->totalDgsInFile;

	char *pRexmtLine = (char *)malloc(sizeof(char)*150000);
	int numRexmt = 0;
	int temp, i;
	
	while(*pbNotDone)
	{	
		pthread_mutex_lock(&mutexRexmt);

		for(i = 0; i < totalDgsInFile && (8 + numRexmt*sizeof(int)) < 150000; i++)
		{
			if(pDroppedDgList[i])
			{
				temp = htonl(i);
				memcpy(pRexmtLine + 8 + numRexmt*sizeof(int), &temp, sizeof(int));
				numRexmt++;
			}
		}
	 	
		pthread_mutex_unlock(&mutexRexmt);
		
	 	if(numRexmt != 0) 
	 	{              	
			sprintf(pRexmtLine, "RXM\n");

			numRexmt = htonl(numRexmt);
			memcpy(pRexmtLine+4, &numRexmt, 4);
			numRexmt = ntohl(numRexmt);

			writen(connfd_tcp, pRexmtLine, 8 + 4 * numRexmt);		
		}

		*pTotalDgsRxmt += numRexmt;
		
		usleep(REXMT_USLEEP);

		numRexmt = 0;
	}
	printf("\nEXITING REXMT THREAD!\n");
	fflush(stdout);
}
void *DiskThread(void *arg)
{
	DiskThreadInfo *pDiskThreadInfo = (DiskThreadInfo *)arg;
	int dgSize = (int)pDiskThreadInfo->dgSize;
	FILE *fpData = (FILE *)pDiskThreadInfo->fpData;
	int totalDgsInFile = (int)pDiskThreadInfo->totalDgsInFile;
	bool *bDiskSleep = (bool *)pDiskThreadInfo->bDiskSleep;
	bool *bAllSent = (bool *)pDiskThreadInfo->bAllSent;
	char **ppDatagrams = (char **)pDiskThreadInfo->ppDatagrams;
	int i = 0, j = 0, k = 0;
	
	int size = 1000; 
	
	char *buff = (char *)malloc(sizeof(char)*(size*(dgSize-4)+1));
	long timeElapsed = 0;


	int fd = fileno(fpData);

	pthread_mutex_lock(&mutexDisk);
	i = glb_diskCount;
	pthread_mutex_unlock(&mutexDisk);
	while(i < totalDgsInFile)
	{
		if(!*bDiskSleep)
		{
			//printf("\nhow many dgs: %d", i);
			printf("Finished transfer! Writing rest of buffer to disk...\n");
			fflush(stdout);
			//pthread_mutex_unlock(&mutexDisk);
			//break;
			*bDiskSleep = true;
		}


		pthread_mutex_lock(&mutexDisk);
		i = glb_diskCount;

		for(k = 0; k+i < totalDgsInFile && k < size && ppDatagrams[k+i] != NULL; k++)
		{
			;
		}

		glb_diskCount += k;
		
		pthread_mutex_unlock(&mutexDisk);
		
		/*if(k != size && k+i != totalDgsInFile)
		{
			//continue;
		}*/

		if(k == 0)
			continue;
		
		//for(j = 0; j+i < totalDgsInFile && j < size && ppDatagrams[j+i] != NULL; j++)
		for(j = 0; j < k; j++)
		{
			memcpy(buff + j*(dgSize-4), ppDatagrams[j+i]+4, dgSize-4);

		}
		//pthread_mutex_unlock(&mutexDisk2);
		fwrite(buff, 1, k*(dgSize-4), fpData);
		//pthread_mutex_unlock(&mutexDisk2);
		fsync(fd);
		
		//i += k;

	}
	
	usleep(2000);

	printf("\nReturning from disk Thread! Closing file!\n");
	printf("%d datagrams written\n", glb_diskCount);

	fflush(stdout);

	return NULL;
}
void *FileProcessingThread(void *arg)
{
	FileInfo *pInputFileInfo = (FileInfo *)arg, curFileInfo;
	FILE *fpInput, *fpOutput;
	char *pCmdLine, *pDg;
	int i, dgSeqNo, dataSize, lastDgSize, numRead, *pIndexListInDgOrder, indexOfLastDg;
	int locationOfDg;
	
	//pthread_detach(pthread_self());
	printf("\nStart post processing file %s ...\n", pInputFileInfo->pFileName);
	fflush(stdout);

	//make a complete copy of the file info input
	curFileInfo.dgSize = pInputFileInfo->dgSize;
	curFileInfo.fileSize = pInputFileInfo->fileSize;
	curFileInfo.totalDgsInFile = pInputFileInfo->totalDgsInFile;
	curFileInfo.myHeaderSize = pInputFileInfo->myHeaderSize;

	strncpy(curFileInfo.pFileName, pInputFileInfo->pFileName, strlen(pInputFileInfo->pFileName));
	curFileInfo.pFileName[strlen(pInputFileInfo->pFileName)] = '\0';
	strncpy(curFileInfo.pTempFileName, pInputFileInfo->pTempFileName, strlen(pInputFileInfo->pTempFileName));
	curFileInfo.pTempFileName[strlen(pInputFileInfo->pTempFileName)] = '\0';
	curFileInfo.pDgSeqNoList = (int *)malloc(sizeof(int)*pInputFileInfo->totalDgsInFile);
	memcpy(curFileInfo.pDgSeqNoList, pInputFileInfo->pDgSeqNoList, (pInputFileInfo->totalDgsInFile)*(pInputFileInfo->myHeaderSize));

	free(pInputFileInfo->pDgSeqNoList);
	free(pInputFileInfo);


	dataSize = curFileInfo.dgSize - curFileInfo.myHeaderSize;
	lastDgSize = (int)(curFileInfo.fileSize - (curFileInfo.totalDgsInFile-1)*dataSize);

	pDg = (char *)malloc(dataSize*sizeof(char));

	if((fpInput = fopen64(curFileInfo.pTempFileName, "r")) == NULL)
	{
		printf("\nError: cannot open temp file: %s\n\n", curFileInfo.pTempFileName);
		exit(0);
	}

	int dgInPlaceCount = 0;
	
	char *temp = (char *)malloc(sizeof(char)*curFileInfo.dgSize);
	int tempIndex = 0;
	
	char *temp2 = (char *)malloc(sizeof(char)*curFileInfo.dgSize);
	i = 0;
	while(dgInPlaceCount < curFileInfo.totalDgsInFile)
	{
		//while in order, increment index
		while(i == curFileInfo.pDgSeqNoList[i])
		{
			i++;
			dgInPlaceCount++; //causes problems (?)
		}
		printf("%4.1f%% ", 100.0*dgInPlaceCount/curFileInfo.totalDgsInFile);
		
		if(dgInPlaceCount >= curFileInfo.totalDgsInFile)
			break;
		if(pDg == NULL || fpOutput == NULL)
		{
			printf("DG or FILE POINTER is NULL!\n");
		}
		//out of order, put the datagram at seqNoList into temp
		locationOfDg = curFileInfo.pDgSeqNoList[i];
		tempIndex = curFileInfo.pDgSeqNoList[locationOfDg];
		seekAndRead(curFileInfo, fpInput, dataSize, lastDgSize, temp, locationOfDg);

		//now put the current datagram into the spot where temp was
		seekAndRead(curFileInfo, fpInput, dataSize, lastDgSize, pDg, i);
		fseek(fpInput, (INT64)locationOfDg*dataSize, SEEK_SET);

		if(numRead != fwrite(pDg, 1, numRead, fpOutput))
		{
			printf("\nError: cannot write to the data file.\n\n");
			exit(0);
		}
		
		//now mark the current spot as empty
		curFileInfo.pDgSeqNoList[locationOfDg] = locationOfDg;
		curFileInfo.pDgSeqNoList[i] = -1;

		while(temp != NULL)
		{
			//put out temp2 from temp's correct position
			if(curFileInfo.pDgSeqNoList[tempIndex] != -1)
			{
				locationOfDg = tempIndex;
				tempIndex = curFileInfo.pDgSeqNoList[locationOfDg];
				seekAndRead(curFileInfo, fpInput, dataSize, lastDgSize, temp2, locationOfDg);
				//put temp into place
				fseek(fpInput, (INT64)locationOfDg*dataSize, SEEK_SET);
				if(numRead != fwrite(temp, 1, numRead, fpOutput))
				{
					printf("\nError: cannot write to the data file.\n\n");
					exit(0);
				}
				curFileInfo.pDgSeqNoList[locationOfDg] = locationOfDg;
				
			}
			//if there is no temp2 to pull out, make temp2 null
			else
			{
				//temp2 = NULL;
				fseek(fpInput, (INT64)tempIndex*dataSize, SEEK_SET);
				if(numRead != fwrite(temp, 1, numRead, fpOutput))
				{
					printf("\nError: cannot write to the data file.\n\n");
					exit(0);
				}
				curFileInfo.pDgSeqNoList[tempIndex] = tempIndex;
				break;
			}
			//assign temp2 to temp, repeat
			temp = temp2;
		}
	}


	fclose(fpInput);
	//fclose(fpOutput);


	printf("\nCompleted post processing file %s.\n\n", curFileInfo.pFileName);
	fflush(stdout);

	pCmdLine = (char *)malloc(200*sizeof(char));
	sprintf(pCmdLine, "mv --force %s %s", curFileInfo.pTempFileName, curFileInfo.pFileName);
	system(pCmdLine);		 

	return NULL;
}

void seekAndRead(FileInfo fileInfo, FILE *fpInput, int dataSize, int lastDgSize, char *temp, int locationOfDg)
{
	int numRead;
	fseek(fpInput, (INT64)locationOfDg*dataSize, SEEK_SET);
	if(locationOfDg == fileInfo.totalDgsInFile - 1)
		numRead = fread(temp, 1, lastDgSize, fpInput);
	else
		numRead = fread(temp, 1, dataSize, fpInput);
	//check for numRead here
	//return NULL;
}


/*void *FileProcessingThread(void *arg)
{
	FileInfo *pInputFileInfo = (FileInfo *)arg, curFileInfo;
	FILE *fpInput, *fpOutput;
	char *pCmdLine, *pDg;
	int i, dgSeqNo, dataSize, lastDgSize, numRead, *pIndexListInDgOrder, locationOfDg, indexOfLastDg;
	
	//pthread_detach(pthread_self());
	printf("\nStart post processing file %s ...\n", pInputFileInfo->pFileName);
	fflush(stdout);

	//make a complete copy of the file info input
	curFileInfo.dgSize = pInputFileInfo->dgSize;
	curFileInfo.fileSize = pInputFileInfo->fileSize;
	curFileInfo.totalDgsInFile = pInputFileInfo->totalDgsInFile;
	curFileInfo.myHeaderSize = pInputFileInfo->myHeaderSize;

	strncpy(curFileInfo.pFileName, pInputFileInfo->pFileName, strlen(pInputFileInfo->pFileName));
	curFileInfo.pFileName[strlen(pInputFileInfo->pFileName)] = '\0';
	strncpy(curFileInfo.pTempFileName, pInputFileInfo->pTempFileName, strlen(pInputFileInfo->pTempFileName));
	curFileInfo.pTempFileName[strlen(pInputFileInfo->pTempFileName)] = '\0';
	curFileInfo.pDgSeqNoList = (int *)malloc(sizeof(int)*pInputFileInfo->totalDgsInFile);
	memcpy(curFileInfo.pDgSeqNoList, pInputFileInfo->pDgSeqNoList, (pInputFileInfo->totalDgsInFile)*(pInputFileInfo->myHeaderSize));
	free(pInputFileInfo->pDgSeqNoList);
	free(pInputFileInfo);

	if((fpInput = fopen64(curFileInfo.pTempFileName, "r")) == NULL)
	{
		printf("\nError: cannot open temp file: %s\n\n", curFileInfo.pTempFileName);
		exit(0);
	}

	if((fpOutput = fopen64(curFileInfo.pFileName, "w")) == NULL)
	{
		printf("\nError: cannot create data file: %s\n\n", curFileInfo.pFileName);
		exit(0);
	}

	dataSize = curFileInfo.dgSize - curFileInfo.myHeaderSize;
	lastDgSize = (int)(curFileInfo.fileSize - (curFileInfo.totalDgsInFile-1)*dataSize);

	pDg = (char *)malloc(sizeof(char)*dataSize);
	pIndexListInDgOrder = (int *)malloc(sizeof(int)*curFileInfo.totalDgsInFile);
	for(i = 0; i < curFileInfo.totalDgsInFile; i ++)
	{
		dgSeqNo = curFileInfo.pDgSeqNoList[i];
		pIndexListInDgOrder[dgSeqNo] = i;
		if(dgSeqNo == curFileInfo.totalDgsInFile - 1) //this is the last dg
			indexOfLastDg = i;
	}

	for(i = 0; i < curFileInfo.totalDgsInFile; i ++)
	{
		locationOfDg = pIndexListInDgOrder[i];
		if(locationOfDg <= indexOfLastDg)
			fseek(fpInput, (INT64)locationOfDg*dataSize, SEEK_SET);
		else
			fseek(fpInput, (INT64)(locationOfDg-1)*dataSize+lastDgSize, SEEK_SET);

		if(i == curFileInfo.totalDgsInFile - 1)
		{
			numRead = fread(pDg, 1, lastDgSize, fpInput);
		}
		else
			numRead = fread(pDg, 1, dataSize, fpInput);

		if(numRead != fwrite(pDg, 1, numRead, fpOutput))
		{
			printf("\nError: cannot write to the data file.\n\n");
			exit(0);
		}
	}

	free(pDg);
	free(pIndexListInDgOrder);
	free(curFileInfo.pDgSeqNoList);
	fclose(fpInput);
	fclose(fpOutput);
	pCmdLine = (char *)malloc(sizeof(char)*200);
	sprintf(pCmdLine, "rm %s", curFileInfo.pTempFileName);
	system(pCmdLine);
	free(pCmdLine);

	printf("\nCompleted post processing file %s.\n\n", curFileInfo.pFileName);
	fflush(stdout);
	return NULL;
}*/



void ThreeWayHandshakeAtReceiver(int sockfd_udp, int dgSize, int *pcwin, int *prate, char *pFileName, INT64 *pFileSize, int *pTotalNumOfDgs, double *prtt)
{
	struct sockaddr_in	cliaddr;
	socklen_t			clilen = sizeof(cliaddr);
    bool				bSYNRecved = false, bACKSYNAcked = false;
    struct timeval 		tv1, tv2;
	fd_set 				readfdset;
	char				*pDg = (char *)malloc(sizeof(char)*dgSize);
	int					rate;

	//printf("\n\n********************************");
	//printf("\nUDP Threeway Handshake @ Receiver:");	
	//fflush(stdout);
	while(!bSYNRecved)
	{
		tv1.tv_sec = 2;
		tv1.tv_usec = 0;
		FD_ZERO(&readfdset);
		FD_SET(sockfd_udp, &readfdset);
	
		if(select(sockfd_udp+1, &readfdset, NULL, NULL, &tv1) > 0)	
		{
			MyRecvFrom(sockfd_udp, pDg, 200, 0, (struct sockaddr *)&cliaddr, &clilen);
			if(strncmp(pDg, "SYN:", 4) == 0)
			{
				bSYNRecved = true;
				sscanf(pDg, "SYN:%d %d %s %lld %d", pcwin, prate, pFileName, pFileSize, pTotalNumOfDgs);
				//printf("\n<===Received SYN (SYN_RCVD):");

				while(!bACKSYNAcked)
				{
					bzero(pDg, sizeof(pDg));

					//*prate = MIN3(MAX_BANDWIDTH, *prate, getMaxSystemBW(*pFileSize,*prate));
					*prate = MIN(MAX_BANDWIDTH, *prate);
					sprintf(pDg, "ACKSYN %d\n", *prate);
					printf("ACKSYN %d\n", *prate);
					sendto(sockfd_udp, pDg, 20, 0, (struct sockaddr *)&cliaddr, clilen);
					//printf("\n===>Sent ACKSYN to client ...");
					tv2.tv_sec = 2;
					tv2.tv_usec = 0;
					FD_ZERO(&readfdset);
					FD_SET(sockfd_udp, &readfdset);
					if(select(sockfd_udp+1, &readfdset, NULL, NULL, &tv2) > 0)
					{
						MyRecvFrom(sockfd_udp, pDg, 20, 0, (struct sockaddr *)&cliaddr, &clilen);
						if(strncmp(pDg, "ACK", 3) == 0)
						{
							sscanf(pDg, "ACK %lf", prtt);
							printf("\n<===Received ACK from client (RTT: %lf ms).", *prtt);
							//printf("\nThreeway Handshake: ESTABLISHED.");
							//printf("\n********************************\n\n");
							bACKSYNAcked = true;
						}
					}	
				}
			}
		}
	} //end of while

	/*printf("\nFile transfer info:");
	printf("\n---------------------------");
	printf("\nMTU size: %d bytes", dgSize + 28);
	printf("\nRequested sending rate: %8.2lf Mbps", ((*pcwin)*dgSize*8.0)/(*psleep+(*pcwin)*dgSize*8.0/BANDWIDTH));//Mbps
	printf("\nFixed cwin: %d datagrams", *pcwin);
	printf("\nFixed sleep_us: %d us", *psleep);
	printf("\nFile size: %lld bytes", *pFileSize);
	printf("\nFile name: %s", pFileName);
	printf("\n---------------------------\n");*/

}

/* change default recv buffer */
int ModifyUDPSystemBuffer(int size, int sockfd_udp/*, socklen_t clilen*/)
{
	void *buffSize;
	char *sysString = (char *)malloc(sizeof(char)*MAX_STRING_LEN);

	struct sockaddr_in cli;
	socklen_t clilen = sizeof(cli);
	
	/* for some reason, Linux/Unix will appropriate x2 as much as asked */
	int newBuffSize = size/2; 
	int actualBuffSize;
	
	/* first set kernel's max buffer size limitation */
	sprintf(sysString, "sysctl -w net.core.rmem_max=%d", newBuffSize);
	system(sysString);
	
	/* check original buffer size */
	getsockopt(sockfd_udp, SOL_SOCKET, SO_RCVBUF, &buffSize, &clilen);
	printf("\n---------\noriginal kernel buffer: %d", buffSize);

	/* set new buffer size */
	setsockopt(sockfd_udp, SOL_SOCKET, SO_RCVBUF, &newBuffSize, sizeof(int));

	/* make sure the system call worked */
	getsockopt(sockfd_udp, SOL_SOCKET, SO_RCVBUF, &buffSize, &clilen);
	printf("\nnew buffer size: %d\n---------\n\n", buffSize);

	/* don't know why this is necessary, maybe someone who is
	 * better at C could tell me a better way to do it */
	sprintf(sysString, "%d", buffSize);
	sscanf(sysString, "%d", &actualBuffSize);

	return actualBuffSize; 
}

void SetThreadPriority(int prio, pthread_attr_t *tattr)
{
	int ret;
	struct sched_param param;

	/* initialized with default attributes */
	ret = pthread_attr_init (tattr);

	/* safe to get existing scheduling param */
	ret = pthread_attr_getschedparam (tattr, &param);

	/* set the priority; others are unchanged */
	param.sched_priority = prio;

	/* setting the new scheduling param */
	ret = pthread_attr_setschedparam (tattr, &param);
}

/*int getMaxSystemBW(INT64 fileSize_bytes, int rateRequest)
{
	INT64 maxFileSize_bytes;
	int maxRate;

	if(fileSize_bytes <= (INT64)(MAX_MEMORY*ONE_MILLION) || MAX_DISK_RATE >= rateRequest)
		return rateRequest;

	//maxFileSize_bytes = (INT64)( MAX_MEMORY / (1 - (MAX_DISK_RATE / rateRequest))*ONE_MILLION);
	//if(fileSize_bytes <= maxFileSize_bytes)
	//	return rateRequest;

	//maxRate = (int)((fileSize_bytes*(MAX_DISK_RATE/8.0)) / (fileSize_bytes-MAX_MEMORY*ONE_MILLION)* 8);
	//return MIN(rateRequest,maxRate);
	
	return (int)((fileSize_bytes*(MAX_DISK_RATE/8.0)) / (fileSize_bytes-MAX_MEMORY*ONE_MILLION)* 8);
}*/

#ifdef LATENCY_SIM
void writenWithLatency(int connfd_tcp, char *toWrite, int length)
{
    struct timeval timestamp;
    gettimeofday(&timestamp, NULL);
	
    //need check here to make sure all globals have been instantiated
    
    datagramLatency[glb_index] = toWrite;
    dgLengthLatency[glb_index] = length;

    timestampLatency[glb_index++] = timestamp;
    if(glb_index > MAX_LATENCY_SIZE)
         glb_index = 0;

    //a thread meanwhile checks the data structure and sends data
    //one it has been cooking for longer than the latency
}


void *writeLatencyThread(void *arg)
{
	int *connfd_tcp = (int *)arg;
	int conn = *connfd_tcp;
	struct timeval timeNow;
	int index = 0;
	int i;
	struct timeval timestamp;

	while(*connfd_tcp != -1)
	{
		if(dgLengthLatency[index] == -1)
		{
		    usleep(1);
		    continue;
		}
		timestamp = timestampLatency[index];
		gettimeofday(&timeNow, NULL);
		if((timeNow.tv_usec-timestamp.tv_usec) + (timeNow.tv_sec-timestamp.tv_sec)* 1000000 > latency)
		{
		    //printf("%d \n", (timeNow.tv_usec-timestamp.tv_usec) + (timeNow.tv_sec-timestamp.tv_sec)* 1000000);
			writen(conn, datagramLatency[index], dgLengthLatency[index]);
		    dgLengthLatency[index] = -1;
		    index++;
		    if(index > MAX_LATENCY_SIZE)
		    	index = 0;
		}
	}
}
#endif

