//Developer:	Qishi Wu
//Date:			June 2004
//Place:		CSMD @ ORNL

#ifndef _UDPSERVER_H_
#define _UDPSERVER_H_

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <stdbool.h>

#include "ioutil.h"

/*******************************************
 * 	Global definitions
 ******************************************/
 
/* Default port. */
#define SERVER_PORT		7058 

/* Max allowable BW in Mbps. */
#define MAX_BANDWIDTH	1000

/* Guess for max disk write rate in Mbps. */
//#define MAX_DISK_RATE	315	

/* Characters for various file names. */
#define MAX_STRING_LEN	100 

/* Bytes for udp kernel recv buffer (keep 8MB+) */
#define	RECV_BUF_SIZE	16777216 

/* Sleep in microseconds for bw control thread. */
#define	BW_CONTROL_USLEEP	1000000 

/* Megabytes of memory allowable for program buffer (suggestion, not hard rule). */
#define	MAX_MEMORY	750 

/* Max allowable fluctuation for rate (Mbps) on control thread. */
//#define	RATE_MAX_FLUCT	50 

/* Number of epochs before trying target throughput again */
//#define RATE_NUM_RETRY 2

/* Sleep in usecs for rexmt thread. */
#define	REXMT_USLEEP	700000 

/* Number of packets to write at a time. */
#define	DISK_PACKET_BLOCK_SIZE	1000 

/* Request retransmission after a period of NUMRTTFORRXMT RTTs. */
//#define NUMRTTFORRXMT 5	

/* Size in bytes of UDP header */
#define HEADER_SIZE 28

#define MAX3(A,B,C) ((A > B) ? ((A > C) ? A : C) : ((B > C) ? B : C))
#define MIN3(A,B,C) ((A < B) ? ((A < C) ? A : C) : ((B < C) ? B : C))
#define MAX(A,B) ((A > B) ? A : B) 
#define MIN(A,B) ((A < B) ? A : B) 

/*******************************************
 * 	Struct declarations
 ******************************************/

/* Parameters for rexmt thread. */
typedef struct RexmtThreadInfo
{
	int		connfd_tcp;
	int		*pTotalDgsRxmt;
	bool	*pbAllSent;
	int		totalDgsInFile;
	int		*pDroppedNum;
	bool	*pbNotDone;
}RexmtThreadInfo;

/* Parameters for disk thread. */
typedef struct DiskThreadInfo
{
	int		dgSize;
	int		totalDgsInFile;
	FILE	*fpData;
	bool	*bDiskSleep;
	bool	*bAllSent;
	char	**ppDatagrams;
}DiskThreadInfo;

/* Parameters for file processing thread. */
typedef struct FileInfo
{
	INT64	fileSize;
	int		myHeaderSize;
	int		dgSize;
	int		totalDgsInFile;
	char	pFileName[MAX_STRING_LEN];
	char	pTempFileName[MAX_STRING_LEN];
	int		*pDgSeqNoList;
}FileInfo;

/* Parameters for rate control thread. */
typedef struct RateControlThreadInfo
{
	int sleep;
	int dgSize;
	int targetRate;
	int connfd_tcp;
	INT64 fileSize;
	bool *bAllSent;
	bool *bFirstDg;
}RateControlThreadInfo;

/*******************************************
 * 	Method prototypes
 ******************************************/

/* Handles passing of parameters at connection. */
void ThreeWayHandshakeAtReceiver(int sockfd_udp, int dgSize, int *pcwin, int *psleep, char *pFileName, INT64 *pFileSize, int *pTotalNumOfDgs, double *prtt);

/* Modifies kernel's default UDP receive buffer */
int ModifyUDPSystemBuffer(int size, int sockfd_udp/*, socklen_t clilen*/);

void SetThreadPriority(int prio, pthread_attr_t *tattr);

int getSleepFromRate(double dgSize, double reqBW);

int getRateFromSleep(double dgSize, double sleep_us);

int getMaxSystemBW(INT64 fileSize_bytes, int rateRequest);

void seekAndRead(FileInfo fileInfo, FILE *fpInput, int dataSize, int lastDgSize, char *temp, int locationOfDg);

/* Takes care of packet retransmission. */
void *RexmtThread(void *arg);

/* Handles writing to disk (out of order). */
void *DiskThread(void *arg);

/* For reconstructing file. */ 
void *FileProcessingThread(void *arg);

/* Dynamic rate control during transfer. */
void *RateControlThread(void *arg);

void writenWithLatency(int connfd_tcp, char *toWrite, int length);

void *writeLatencyThread(void *arg);

int startTransfer(int sockfd_udp, int connfd_tcp);

#endif

